//@line 36 "/builds/tinderbox/Sb-Trunk-l10n/Linux_2.4.21-32.0.1.EL_Clobber/mozilla/calendar/locales/../../../l10n/fr/calendar/sunbird-l10n.js"

//@line 38 "/builds/tinderbox/Sb-Trunk-l10n/Linux_2.4.21-32.0.1.EL_Clobber/mozilla/calendar/locales/../../../l10n/fr/calendar/sunbird-l10n.js"

//@line 41 "/builds/tinderbox/Sb-Trunk-l10n/Linux_2.4.21-32.0.1.EL_Clobber/mozilla/calendar/locales/../../../l10n/fr/calendar/sunbird-l10n.js"
pref("calendar.week.start", 1);

//@line 44 "/builds/tinderbox/Sb-Trunk-l10n/Linux_2.4.21-32.0.1.EL_Clobber/mozilla/calendar/locales/../../../l10n/fr/calendar/sunbird-l10n.js"
pref("calendar.week.d0sundaysoff", true);
pref("calendar.week.d1mondaysoff", false);
pref("calendar.week.d2tuesdaysoff", false);
pref("calendar.week.d3wednesdaysoff", false);
pref("calendar.week.d4thursdaysoff", false);
pref("calendar.week.d5fridaysoff", false);
pref("calendar.week.d6saturdaysoff", true);

pref("general.useragent.locale", "fr");

//@line 55 "/builds/tinderbox/Sb-Trunk-l10n/Linux_2.4.21-32.0.1.EL_Clobber/mozilla/calendar/locales/../../../l10n/fr/calendar/sunbird-l10n.js"
pref("calendar.categories.names", "Anniversaire de mariage,Anniversaire,Affaire,Appels,Clients,Compétitions,Consommation,Préférés,Suivis,Cadeaux,Congés,Idées,Problèmes,Divers,Personnel,Projets,Jour férié,Statut,Fournisseurs,Voyages,Vacances");
